package bwmonitor

import (
	"bufio"
	"errors"
	"os"
	"regexp"
	"strconv"
	"strings"
	"time"
)

var spacesRx = regexp.MustCompile(`\s+`)

type netDevCounts struct {
	bytesSent int
}

func getBytesSentForDevice(dev string) (uint64, error) {
	file, err := os.Open("/proc/net/dev")
	if err != nil {
		return 0, err
	}
	defer file.Close()

	input := bufio.NewScanner(file)
	for input.Scan() {
		line := spacesRx.Split(strings.TrimSpace(input.Text()), -1)
		curDev := line[0][0 : len(line[0])-1]
		if curDev == dev {
			bytes, err := strconv.ParseUint(line[9], 10, 64)
			if err != nil {
				return 0, err
			}
			return bytes, nil
		}
	}
	return 0, errors.New("device not found")
}

type BandwidthMonitor struct {
	device  string
	counter uint64
	stamp   time.Time
	period  time.Duration
	stop    chan bool
	rate    float64
}

func NewBandwidthMonitor(dev string) *BandwidthMonitor {
	bw := &BandwidthMonitor{
		device: dev,
		stamp:  time.Now(),
		period: 30 * time.Second,
		stop:   make(chan bool),
	}
	go bw.run()
	return bw
}

func (bw *BandwidthMonitor) Close() {
	close(bw.stop)
}

func (bw *BandwidthMonitor) GetRate() float64 {
	return bw.rate
}

func (bw *BandwidthMonitor) run() {
	t := time.NewTicker(bw.period)
	for {
		select {
		case <-t.C:
			if c, err := getBytesSentForDevice(bw.device); err == nil {
				now := time.Now()
				bw.rate = float64(c - bw.counter) / now.Sub(bw.stamp).Seconds()
				bw.counter = c
				bw.stamp = now
			}
		case <-bw.stop:
			return
		}
	}
}

type BandwidthUsageMonitor struct {
	*BandwidthMonitor

	bwLimit float64
}

func NewBandwidthUsageMonitor(dev string, bwLimit float64) *BandwidthUsageMonitor {
	return &BandwidthUsageMonitor{
		NewBandwidthMonitor(dev),
		bwLimit,
	}
}

func (bu *BandwidthUsageMonitor) GetUsage() float64 {
	return bu.GetRate() / bu.bwLimit
}
