package node

import (
	"encoding/xml"
	"errors"
	"fmt"
	"io"
	"log"
	"net"
	"net/http"
	"os"
	"os/exec"
	"strconv"
	"time"

	"git.autistici.org/ale/autoradio"
	"git.autistici.org/ale/autoradio/instrumentation"
)

var (
	statusPage        = "/status-autoradio.xsl"
	icecastConfigFile = "/etc/icecast2/icecast.xml"
	icecastReloadCmd  = "/usr/sbin/service icecast2 reload || /usr/sbin/service icecast2 start"

	icecastOk = instrumentation.NewGauge("icecast.ok")
)

// Icecast returns empty fields in our status handler, which we'll
// need to turn into integers (the xml unmarshaler will return an
// error in this specific case), so we use a separate type for
// decoding the status page output. This would be much simpler if I
// knew how to get the XSLT to put a default value in the output
// instead of an empty field...
type icecastMountStatusUnparsed struct {
	Name         string `xml:"name,attr"`
	Listeners    string `xml:"listeners"`
	BitRate      string `xml:"bitrate"`
	Quality      string `xml:"quality"`
	VideoQuality string `xml:"video-quality"`
	FrameSize    string `xml:"frame-size"`
	FrameRate    string `xml:"frame-rate"`
}

type icecastStatusUnparsed struct {
	XMLName xml.Name                     `xml:"status"`
	Mounts  []icecastMountStatusUnparsed `xml:"mount"`
}

type IcecastStatus struct {
	Mounts []autoradio.IcecastMountStatus
	Up     bool
}

type IcecastController struct {
	config *icecastConfig
	status *IcecastStatus
	stop   chan bool
}

func NewIcecastController(publicIp string, stop chan bool) *IcecastController {
	return &IcecastController{
		config: newIcecastConfig(publicIp),
		status: &IcecastStatus{},
		stop:   make(chan bool, 1),
	}
}

// Reload the icecast daemon. Redirects output to our standard error
// for debugging purposes.
func (ic *IcecastController) reload() error {
	cmd := exec.Command("/bin/sh", "-c", icecastReloadCmd)
	cmd.Stdout = os.Stderr
	cmd.Stderr = os.Stderr
	return cmd.Run()
}

// Kill sources connected to local streams.
func (ic *IcecastController) killSources(conf *clusterConfig) error {
	var anyErr error
	client := &http.Client{}
	for _, m := range conf.ListMounts() {
		req, err := http.NewRequest("GET", fmt.Sprintf("http://localhost:%d/admin/killsource?mount=%s", autoradio.IcecastPort, autoradio.MountNameToIcecastPath(m.Name)), nil)
		if err != nil {
			anyErr = err
			continue
		}
		req.SetBasicAuth("admin", getIcecastAdminPassword())
		resp, err := client.Do(req)
		if err != nil {
			anyErr = err
			continue
		}
		resp.Body.Close()
		if resp.StatusCode != 200 {
			anyErr = fmt.Errorf("HTTP status %s", resp.Status)
		}
	}
	return anyErr
}

// Update reloads the Icecast daemon with a new configuration.
func (ic *IcecastController) Update(conf *clusterConfig, isMaster bool, masterAddr net.IP) error {
	if !isMaster && masterAddr == nil {
		return errors.New("unknown system state")
	}

	// Try to kill sources connected to the local icecast daemon
	// before reloading, otherwise we'll have problems on master
	// -> slave transitions (for example, on a restart of radiod)
	// as sources will be "stuck" preventing the new configuration
	// from taking effect.
	if err := ic.killSources(conf); err != nil {
		log.Printf("error killing sources: %v", err)
	}

	// Write a new configuration (atomically).
	ic.config.Update(conf, isMaster, masterAddr.String())
	tmpf := icecastConfigFile + ".tmp"
	defer os.Remove(tmpf)
	if err := ic.config.EncodeToFile(tmpf); err != nil {
		return err
	}
	if err := os.Rename(tmpf, icecastConfigFile); err != nil {
		return err
	}

	// Tell the Icecast daemon to reload its configuration.
	return ic.reload()
}

func (ic *IcecastController) GetStatus() *IcecastStatus {
	return ic.status
}

func (ic *IcecastController) statusUpdater() {
	t := time.NewTicker(3 * time.Second)
	downStatus := &IcecastStatus{}
	for {
		select {
		case <-t.C:
			if status, err := ic.fetchStatus(); err == nil {
				ic.status = status
				icecastOk.Set(1)
			} else {
				log.Printf("error checking iceast status: %v", err)
				ic.status = downStatus
				icecastOk.Set(0)
			}
		case <-ic.stop:
			return
		}
	}
}

func (ic *IcecastController) fetchStatus() (*IcecastStatus, error) {
	resp, err := http.Get(fmt.Sprintf("http://localhost:%d%s", autoradio.IcecastPort, statusPage))
	if err != nil {
		return nil, err
	}
	defer resp.Body.Close()
	return ic.parseStatusPage(resp.Body)
}

func (ic *IcecastController) parseStatusPage(input io.Reader) (*IcecastStatus, error) {
	var ustatus icecastStatusUnparsed
	if err := xml.NewDecoder(input).Decode(&ustatus); err != nil {
		return nil, err
	}

	// Quick converters from string that default to 0.
	toi := func(s string) int {
		if i, err := strconv.Atoi(s); err == nil {
			return i
		}
		return 0
	}
	tof := func(s string) float64 {
		if f, err := strconv.ParseFloat(s, 64); err == nil {
			return f
		}
		return 0
	}

	status := IcecastStatus{
		Up:     true,
		Mounts: make([]autoradio.IcecastMountStatus, 0, len(ustatus.Mounts)),
	}
	for _, um := range ustatus.Mounts {
		m := autoradio.IcecastMountStatus{
			Name:         autoradio.IcecastPathToMountName(um.Name),
			Listeners:    toi(um.Listeners),
			BitRate:      toi(um.BitRate),
			Quality:      tof(um.Quality),
			VideoQuality: tof(um.VideoQuality),
			FrameSize:    um.FrameSize,
			FrameRate:    tof(um.FrameRate),
		}
		status.Mounts = append(status.Mounts, m)
	}

	return &status, nil
}

func (ic *IcecastController) Run() {
	ic.statusUpdater()
}
